import logging

import urwid
from sen.exceptions import NotifyError


logger = logging.getLogger(__name__)


class WidgetBase(urwid.ListBox):
    """
    common class fot widgets
    """

    def __init__(self, *args, **kwargs):
        self.search_string = None
        self.filter_query = None
        super().__init__(*args, **kwargs)
        self.ro_content = self.body[:]  # unfiltered content of a widget

    def set_body(self, widgets):
        self.body[:] = widgets

    def reload_widget(self):
        # this is the easiest way to refresh body
        self.body[:] = self.body

    def _search(self, reverse_search=False):
        if self.search_string is None:
            raise NotifyError("No search pattern specified.")
        if not self.search_string:
            self.search_string = None
            return
        pos = self.focus_position
        original_position = pos
        wrapped = False
        while True:
            if reverse_search:
                obj, pos = self.body.get_prev(pos)
            else:
                obj, pos = self.body.get_next(pos)
            if obj is None:
                # wrap
                wrapped = True
                if reverse_search:
                    obj, pos = self.body[-1], len(self.body)
                else:
                    obj, pos = self.body[0], 0
            if wrapped and (
                        (pos > original_position and not reverse_search) or
                        (pos < original_position and reverse_search)
            ):
                self.search_string = None
                raise NotifyError("Pattern not found: %r." % self.search_string)
            # FIXME: figure out nicer search api
            if hasattr(obj, "matches_search"):
                condition = obj.matches_search(self.search_string)
            else:
                condition = self.search_string in obj.original_widget.text
            if condition:
                self.set_focus(pos)
                self.reload_widget()
                break

    def filter(self, s, widgets_to_filter=None):
        s = s.strip()

        if not s:
            self.filter_query = None
            self.body[:] = self.ro_content
            return

        widgets = []
        for obj in widgets_to_filter or self.ro_content:

            # FIXME: figure out nicer search api
            if hasattr(obj, "matches_search"):
                condition = obj.matches_search(s)
            else:
                condition = s in obj.original_widget.text

            if condition:
                widgets.append(obj)
        if not widgets_to_filter:
            self.filter_query = s
        return widgets

    def find_previous(self, search_pattern=None):
        if search_pattern is not None:
            self.search_string = search_pattern
        self._search(reverse_search=True)

    def find_next(self, search_pattern=None):
        if search_pattern is not None:
            self.search_string = search_pattern
        self._search()

    def status_bar(self):
        columns_list = []

        def add_subwidget(markup, color_attr=None):
            if color_attr is None:
                w = urwid.AttrMap(urwid.Text(markup), "status_text")
            else:
                w = urwid.AttrMap(urwid.Text(markup), color_attr)
            columns_list.append((len(markup), w))

        if self.search_string:
            add_subwidget("Search: ")
            add_subwidget(repr(self.search_string))

        if self.search_string and self.filter_query:
            add_subwidget(", ")

        if self.filter_query:
            add_subwidget("Filter: ")
            add_subwidget(repr(self.filter_query))

        return columns_list


class VimMovementListBox(WidgetBase):
    """
    ListBox with vim-like movement which can be inherited in other widgets
    """

    def __init__(self, *args, **kwargs):
        # we want "gg"!
        self.cached_key = None
        super().__init__(*args, **kwargs)

    def keypress(self, size, key):
        logger.debug("VimListBox keypress %r", key)

        # FIXME: workaround so we allow "gg" only, and not "g*"
        if self.cached_key == "g" and key != "g":
            self.cached_key = None

        if key == "j":
            return super().keypress(size, "down")
        elif key == "k":
            return super().keypress(size, "up")
        elif key == "ctrl d":
            try:
                self.set_focus(self.get_focus()[1] + 10)
            except IndexError:
                self.set_focus(len(self.body) - 1)
            self.reload_widget()
            return
        elif key == "ctrl u":
            try:
                self.set_focus(self.get_focus()[1] - 10)
            except IndexError:
                self.set_focus(0)
            self.reload_widget()
            return
        elif key == "G":
            self.set_focus(len(self.body) - 1)
            self.body[:] = self.body
            return
        elif key == "g":
            if self.cached_key is None:
                self.cached_key = "g"
            elif self.cached_key == "g":
                self.set_focus(0)
                self.reload_widget()
                self.cached_key = None
            return
        key = super().keypress(size, key)
        return key



import logging
import re
import threading

import urwid
from sen.exceptions import NotifyError

from sen.tui.constants import MAIN_LIST_FOCUS
from sen.docker_backend import DockerImage, DockerContainer
from sen.tui.widgets.list.base import VimMovementListBox
from sen.tui.widgets.list.util import get_map, get_time_attr_map, get_operation_notify_widget
from sen.tui.widgets.util import AdHocAttrMap
from sen.util import log_traceback


logger = logging.getLogger(__name__)


class MainLineWidget(urwid.AttrMap):
    FIRST_COL = 12
    THIRD_COL = 16
    FOURTH_COL = 30

    def __init__(self, o):
        self.docker_object = o
        self.widgets = []
        columns = []

        if isinstance(o, DockerImage):
            image_id = AdHocAttrMap(urwid.Text(o.image_id[:12]), get_map())
            self.widgets.append(image_id)
            columns.append((self.FIRST_COL, image_id))

            command = AdHocAttrMap(urwid.Text(o.command, wrap="clip"),
                                   get_map(defult="main_list_ddg"))
            self.widgets.append(command)
            columns.append(command)

            base_image = o.base_image()
            base_image_text = ""
            if base_image:
                base_image_text = base_image.short_name
            base_image_w = AdHocAttrMap(urwid.Text(base_image_text, wrap="clip"), get_map())
            self.widgets.append(base_image_w)
            columns.append((self.THIRD_COL, base_image_w))

            time = AdHocAttrMap(urwid.Text(o.display_time_created()), get_time_attr_map(o.created))
            self.widgets.append(time)
            columns.append((self.FOURTH_COL, time))

            names_widgets = []

            def add_subwidget(markup, color_attr):
                w = AdHocAttrMap(urwid.Text(markup), get_map(color_attr))
                names_widgets.append((len(markup), w))
                self.widgets.append(w)

            for n in o.names:
                if n.registry:
                    add_subwidget(n.registry + "/", "main_list_dg")
                if n.namespace and n.repo:
                    add_subwidget(n.namespace + "/" + n.repo, "main_list_lg")
                else:
                    if n.repo == "<none>":
                        add_subwidget(n.repo, "main_list_dg")
                    else:
                        add_subwidget(n.repo, "main_list_lg")
                if n.tag:
                    if n.tag not in ["<none>", "latest"]:
                        add_subwidget(":" + n.tag, "main_list_dg")
                add_subwidget(", ", "main_list_dg")
            names_widgets = names_widgets[:-1]
            names = AdHocAttrMap(urwid.Columns(names_widgets), get_map())
            self.widgets.append(names)
            columns.append(names)

        elif isinstance(o, DockerContainer):
            container_id = AdHocAttrMap(urwid.Text(o.container_id[:12]), get_map())
            self.widgets.append(container_id)
            columns.append((12, container_id))

            command = AdHocAttrMap(urwid.Text(o.command, wrap="clip"),
                                   get_map(defult="main_list_ddg"))
            self.widgets.append(command)
            columns.append(command)

            image = AdHocAttrMap(urwid.Text(o.image_name()), get_map())
            self.widgets.append(image)
            columns.append((self.THIRD_COL, image))

            if o.status.startswith("Up"):
                attr_map = get_map("main_list_green")
            elif o.status.startswith("Exited (0)"):
                attr_map = get_map("main_list_orange")
            elif o.status.startswith("Created"):
                attr_map = get_map("main_list_yellow")
            else:
                attr_map = get_map("main_list_red")
            status = AdHocAttrMap(urwid.Text(o.status), attr_map)
            self.widgets.append(status)
            columns.append((self.FOURTH_COL, status))

            name = AdHocAttrMap(urwid.Text(o.short_name, wrap="clip"), get_map())
            self.widgets.append(name)
            columns.append(name)

        self.columns = urwid.Columns(columns, dividechars=1)
        super().__init__(self.columns,
                         "main_list_dg",
                         focus_map=MAIN_LIST_FOCUS)

    def matches_search(self, s):
        return self.docker_object.matches_search(s)

    def selectable(self):
        return True

    def render(self, size, focus=False):
        for w in self.widgets:
            w.set_map('focus' if focus else 'normal')
        return urwid.AttrMap.render(self, size, focus)

    def keypress(self, size, key):
        logger.debug("%r keypress, focus: %s", self, self.focus)
        return key

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.docker_object)

    def __str__(self):
        return "{}".format(self.docker_object)


class MainListBox(VimMovementListBox):
    def __init__(self, docker_backend, ui):
        self.d = docker_backend
        self.ui = ui
        self.walker = urwid.SimpleFocusListWalker([])
        super(MainListBox, self).__init__(self.walker)

        self.thread = threading.Thread(target=self.realtime_updates, daemon=True)
        self.thread.start()

    def populate(self, focus_on_top=False):
        widgets = self._assemble_initial_content()
        self.walker[:] = widgets
        self.ro_content = widgets
        if focus_on_top:
            self.set_focus(0)
        self.ui.refresh()

    def realtime_updates(self):
        """
        update listing realtime as events come from docker

        :return:
        """
        for content in self.d.realtime_updates():
            widgets = []
            for o in content:
                try:
                    line = MainLineWidget(o)
                except Exception as ex:
                    logger.error("%r", ex)
                    continue
                widgets.append(line)
            self.walker[:] = widgets
            self.ui.reload_footer()
            self.ui.refresh()

    def _assemble_initial_content(self):
        def query_notify(operation):
            w = get_operation_notify_widget(operation, display_always=False)
            if w:
                self.ui.notify_widget(w)

        widgets = []
        query, c_op, i_op = self.d.filter()
        for o in query:
            line = MainLineWidget(o)
            widgets.append(line)
        query_notify(i_op)
        query_notify(c_op)
        return widgets

    @property
    def focused_docker_object(self):
        try:
            return self.get_focus()[0].docker_object
        except AttributeError as ex:
            raise NotifyError("Nothing selected!")

    def filter(self, s, widgets_to_filter=None):
        s = s.strip()
        if not s:
            self.filter_query = None
            self.populate(focus_on_top=True)
            return

        backend_query = {
            "cached": True,
            "containers": False,
            "images": False,
        }

        def containers():
            backend_query["containers"] = True

        def images():
            backend_query["images"] = True

        def running():
            backend_query["stopped"] = False

        query_conf = [
            {
                "query_keys": ["t", "type"],
                "query_values": ["c", "container", "containers"],
                "callback": containers
            }, {
                "query_keys": ["t", "type"],
                "query_values": ["i", "images", "images"],
                "callback": images
            }, {
                "query_keys": ["s", "state"],
                "query_values": ["r", "running"],
                "callback": running
            },
        ]
        query_list = re.split(r"[\s,]", s)
        unprocessed = []
        do_query = False
        for query_str in query_list:
            # process here x=y queries and pass rest to parent filter()
            try:
                query_key, query_value = query_str.split("=", 1)
            except ValueError:
                unprocessed.append(query_str)
            else:
                do_query = True
                logger.debug("looking up query key %r and query value %r", query_key, query_value)
                for c in query_conf:
                    if query_key in c["query_keys"] and query_value in c["query_values"]:
                        c["callback"]()
                        break
                else:
                    raise NotifyError("Invalid query string: %r", query_str)
        if do_query:
            widgets = []
            logger.debug("doing query %s", backend_query)
            query, c_op, i_op = self.d.filter(**backend_query)
            for o in query:
                line = MainLineWidget(o)
                widgets.append(line)
            if unprocessed:
                new_query = " ".join(unprocessed)
                logger.debug("doing parent query for unprocessed string: %r", new_query)
                widgets = super().filter(new_query, widgets_to_filter=widgets)
        else:
            logger.debug("doing parent query: %r", s)
            widgets = super().filter(s)
        self.filter_query = s
        return widgets

    def keypress(self, size, key):
        # FIXME: put this into own file
        # these functions will be executed in threads
        # provide arguments, don't access self.<attrib> b/c those will be
        # evaluated when the code is running, not when calling it
        @log_traceback
        def run_and_report_on_fail(fn_name, docker_object, pre_message,
                                   notif_level="info"):
            self.ui.notify_message(pre_message)
            try:
                operation = getattr(docker_object, fn_name)()
            except AttributeError:
                log_txt = "you can't {} {}".format(fn_name, docker_object)
                logger.error(log_txt)
                notif_txt = "You can't {} {} {!r}.".format(
                    fn_name,
                    docker_object.pretty_object_type.lower(),
                    docker_object.short_name)
                self.ui.notify_message(notif_txt, level="error")
            except Exception as ex:
                self.ui.notify_message(str(ex), level="error")
                raise
            else:
                self.ui.remove_notification_message(pre_message)
                self.ui.notify_widget(
                    get_operation_notify_widget(operation, notif_level=notif_level)
                )
                # we don't need to refresh whole buffer here, since we are getting realtime
                # updates using events call
                # self.ui.refresh_main_buffer()

        @log_traceback
        def do_and_report_on_fail(f, docker_object):
            try:
                if docker_object:
                    f(docker_object)
                else:
                    f()
            except NotifyError as ex:
                self.ui.notify_message(str(ex), level="error")
                logger.error(repr(ex))

        logger.debug("size %r, key %r", size, key)

        try:
            if key == "@":
                self.ui.run_in_background(self.populate, None)
            elif key == "i":
                self.ui.run_in_background(do_and_report_on_fail, self.ui.inspect, self.focused_docker_object)
                return
            elif key == "l":
                self.ui.run_in_background(do_and_report_on_fail, self.ui.display_logs, self.focused_docker_object)
                return
            elif key == "f":
                self.ui.run_in_background(do_and_report_on_fail, self.ui.display_and_follow_logs, self.focused_docker_object)
                return
            elif key == "enter":
                self.ui.run_in_background(do_and_report_on_fail, self.ui.display_image_info, self.focused_docker_object)
                return
            elif key == "d":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "remove",
                    self.focused_docker_object,
                    "Removing {} {}...".format(self.focused_docker_object.pretty_object_type.lower(),
                                               self.focused_docker_object.short_name),
                    notif_level="important"
                )
                return
            elif key == "s":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "start",
                    self.focused_docker_object,
                    "Starting container {}...".format(self.focused_docker_object.short_name)
                )
                return
            elif key == "r":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "restart",
                    self.focused_docker_object,
                    "Restarting container {}...".format(self.focused_docker_object.short_name)
                )
                return
            elif key == "t":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "stop",
                    self.focused_docker_object,
                    "Stopping container {}...".format(self.focused_docker_object.short_name)
                )
                return
            elif key == "p":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "pause",
                    self.focused_docker_object,
                    "Pausing container {}...".format(self.focused_docker_object.short_name)
                )
                return
            elif key == "u":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "unpause",
                    self.focused_docker_object,
                    "Unpausing container {}...".format(self.focused_docker_object.short_name)
                )
                return
            elif key == "X":
                self.ui.run_in_background(
                    run_and_report_on_fail,
                    "kill",
                    self.focused_docker_object,
                    "Killing container {}...".format(self.focused_docker_object.short_name)
                )
                return
        except NotifyError as ex:
            self.ui.notify_message(str(ex), level="error")
            logger.error(repr(ex))

        key = super(MainListBox, self).keypress(size, key)
        return key

    def status_bar(self):
        columns_list = []

        def add_subwidget(markup, color_attr=None):
            if color_attr is None:
                w = urwid.AttrMap(urwid.Text(markup), "status_text")
            else:
                w = urwid.AttrMap(urwid.Text(markup), color_attr)
            columns_list.append((len(markup), w))

        add_subwidget("Images: ")
        images_count = len(self.d.get_images(cached=True).response)
        if images_count < 10:
            add_subwidget(str(images_count), "status_text_green")
        elif images_count < 50:
            add_subwidget(str(images_count), "status_text_yellow")
        else:
            add_subwidget(str(images_count), "status_text_orange")

        add_subwidget(", Containers: ")
        containers_count = len(self.d.get_containers(cached=True, stopped=True).response)
        if containers_count < 5:
            add_subwidget(str(containers_count), "status_text_green")
        elif containers_count < 30:
            add_subwidget(str(containers_count), "status_text_yellow")
        elif containers_count < 100:
            add_subwidget(str(containers_count), "status_text_orange")
        else:
            add_subwidget(str(containers_count), "status_text_red")

        add_subwidget(", Running: ")
        running_containers = self.d.get_containers(cached=True, stopped=False).response
        running_containers_n = len(running_containers)
        add_subwidget(str(running_containers_n),
                      "status_text_green" if running_containers_n > 0 else "status_text")

        parent_cols = super().status_bar()
        if parent_cols:
            add_subwidget(", ")
        return columns_list + parent_cols



import pytest
from flexmock import flexmock
from sen.tui.widgets.list.common import ScrollableListBox, AsyncScrollableListBox
from .constants import SCREEN_WIDTH, SCREEN_HEIGHT


class DataGenerator:
    @classmethod
    def text(cls, prefix="line", lines_no=3, return_bytes=False):
        s = "\n".join(["{}{}".format(prefix, x+1) for x in range(lines_no)])
        if return_bytes:
            return s.encode("utf-8")
        return s

    @classmethod
    def stream(cls, prefix="line", lines_no=3, return_bytes=False):
        response = []
        for x in range(lines_no):
            l = "{}{}".format(prefix, x+1)
            if return_bytes:
                l = l.encode("utf-8")
            response.append(l)
        return iter(response)

    @classmethod
    def render(cls, prefix="line", lines_no=3, return_bytes=True):
        w = "{:%d}" % SCREEN_WIDTH
        response = []
        for x in range(SCREEN_HEIGHT):
            if x >= lines_no:
                l = w.format("")
            else:
                l = w.format("{}{}".format(prefix, x+1))
            if return_bytes:
                response.append(l.encode("utf-8"))
            else:
                response.append(l)
        return response


@pytest.mark.parametrize("inp,expected", [
    (DataGenerator.text(), DataGenerator.render()),
    (DataGenerator.text(return_bytes=True), DataGenerator.render()),
    (DataGenerator.text(prefix="liné"), DataGenerator.render(prefix="liné")),
    (DataGenerator.text(prefix="liné", return_bytes=True), DataGenerator.render(prefix="liné")),
])
def test_scrollable_listbox(inp, expected):
    lb = ScrollableListBox(inp)
    canvas = lb.render((SCREEN_WIDTH, SCREEN_HEIGHT))
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    assert text == expected


@pytest.mark.parametrize("inp,expected", [
    (DataGenerator.stream(), DataGenerator.render()),
    (DataGenerator.stream(return_bytes=True), DataGenerator.render()),
    (DataGenerator.stream(prefix="liné"), DataGenerator.render(prefix="liné")),
    (DataGenerator.stream(prefix="liné", return_bytes=True), DataGenerator.render(prefix="liné")),
])
def test_async_scrollable_listbox(inp, expected):
    ui = flexmock(refresh=lambda: None)
    lb = AsyncScrollableListBox(inp, ui)
    lb.thread.join()
    canvas = lb.render((SCREEN_WIDTH, SCREEN_HEIGHT))
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    assert text == expected



import logging

import urwidtrees

from sen.tui.widgets.info import SelectableText
from sen.tui.widgets.list.util import RowWidget
from sen.tui.widgets.util import get_basic_image_markup

logger = logging.getLogger(__name__)


class TreeNodeWidget(SelectableText):
    def __init__(self, ui, docker_image):
        self.ui = ui
        self.docker_image = docker_image
        super().__init__(get_basic_image_markup(docker_image))

    def keypress(self, size, key):
        logger.debug("%s %s %s", self.__class__, key, size)
        if key == "enter":
            self.ui.display_image_info(self.docker_image)
            return
        return key


class TreeBackend(urwidtrees.Tree):
    def __init__(self, docker_backend, ui):
        super().__init__()
        self.ui = ui
        self.root = docker_backend.scratch_image

    def __getitem__(self, pos):
        return RowWidget([TreeNodeWidget(self.ui, pos)])

    # Tree API
    def parent_position(self, pos):
        return pos.parent_image

    def first_child_position(self, pos):
        ch = pos.children
        if ch:
            return pos.children[0]
        else:
            return None

    def last_child_position(self, pos):
        ch = pos.children
        if ch:
            return pos.children[-1]
        else:
            return None

    def next_sibling_position(self, pos):
        return pos.get_next_sibling()

    def prev_sibling_position(self, pos):
        return pos.get_prev_sibling()


class ImageTree(urwidtrees.TreeBox):
    def __init__(self, docker_backend, ui):
        tree = TreeBackend(docker_backend, ui)

        # We hide the usual arrow tip and use a customized collapse-icon.
        t = urwidtrees.CollapsibleArrowTree(
            tree,
            arrow_att="tree",  # lines, tip
            icon_collapsed_att="tree",  # +
            icon_expanded_att="tree",  # -
            icon_frame_att="tree",  # [ ]
        )
        super().__init__(t)



import copy
import functools
import json
import logging
import datetime
import traceback

from sen.exceptions import TerminateApplication

import docker
import humanize


logger = logging.getLogger(__name__)


class ImageNameStruct(object):
    """
    stolen from atomic-reactor; thanks @mmilata!
    """
    def __init__(self, registry=None, namespace=None, repo=None, tag=None):
        self.registry = registry
        self.namespace = namespace
        self.repo = repo
        self.tag = tag

    @classmethod
    def parse(cls, image_name):
        result = cls()

        # registry.org/namespace/repo:tag
        s = image_name.split('/', 2)

        if len(s) == 2:
            if '.' in s[0] or ':' in s[0]:
                result.registry = s[0]
            else:
                result.namespace = s[0]
        elif len(s) == 3:
            result.registry = s[0]
            result.namespace = s[1]
        if result.namespace == 'library':
            # https://github.com/projectatomic/atomic-reactor/issues/45
            logger.debug("namespace 'library' -> ''")
            result.namespace = None
        result.repo = s[-1]

        try:
            result.repo, result.tag = result.repo.rsplit(':', 1)
        except ValueError:
            pass

        return result

    def to_str(self, registry=True, tag=True, explicit_tag=False,
               explicit_namespace=False):
        if self.repo is None:
            raise RuntimeError('No image repository specified')

        result = self.repo if self.repo != "<none>" else ""

        # don't display <none> junk
        if tag and self.tag and self.tag != "<none>":
            result = '{0}:{1}'.format(result, self.tag)
        elif tag and explicit_tag and self.tag != "<none>":
            result = '{0}:{1}'.format(result, 'latest')

        # don't display <none> junk
        if self.namespace:
            result = '{0}/{1}'.format(self.namespace, result)
        elif explicit_namespace:
            result = '{0}/{1}'.format('library', result)

        if registry and self.registry:
            result = '{0}/{1}'.format(self.registry, result)

        return result

    def __str__(self):
        return self.to_str(registry=True, tag=True)

    def __repr__(self):
        return "ImageName(image=%s)" % repr(self.to_str())

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.to_str())

    def copy(self):
        return ImageNameStruct(
            registry=self.registry,
            namespace=self.namespace,
            repo=self.repo,
            tag=self.tag)


def operation(fmt_str):
    def wrap(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            command_took = None
            pretty_message = ""
            # cached queries do not access backend -- we don't care about that
            if kwargs.get("cached", False) is True:
                response = func(self, *args, **kwargs)
            else:
                before = datetime.datetime.now()
                response = func(self, *args, **kwargs)
                after = datetime.datetime.now()

                # we want milliseconds, not seconds
                command_took = (after - before).total_seconds() * 1000
                logger.debug("%s(%s, %s) %s -> [%f ms]", func.__name__, args, kwargs, self,
                             command_took)
            if fmt_str:
                pretty_message = fmt_str.format(object_type=getattr(self, "pretty_object_type", ""),
                                                object_short_name=getattr(self, "short_name", ""))
            return Operation(response, pretty_message=pretty_message, took=command_took)
        return wrapper
    return wrap


class Operation:
    """
    class for describing performed operation
    """

    def __init__(self, response, pretty_message="", took=None):
        self.response = response
        self.pretty_message = pretty_message
        self.took = took


class DockerObject:
    """
    Common base for images and containers
    """
    def __init__(self, data, docker_backend, object_id=None):
        self._id = object_id
        self._short_id = None
        self.data = data  # `client.containers` or `client.images`
        self.docker_backend = docker_backend
        self._created = None
        self._inspect = None
        self._names = None

    @property
    def d(self):
        """
        shortcut for instance of Docker client
        """
        return self.docker_backend.client

    @property
    def created_int(self):
        return self.data["Created"]

    @property
    def created(self):
        if self._created is None:
            self._created = datetime.datetime.fromtimestamp(self.data["Created"])
        return self._created

    def set_id(self):
        if self._id is None:
            try:
                self._id = self.data["Id"]
            except KeyError:
                raise RuntimeError("initial data not specified")

    @property
    def short_id(self):
        if self._short_id is None:
            self.set_id()
            self._short_id = self._id[:12]
        return self._short_id

    def display_time_created(self):
        return humanize.naturaltime(self.created)

    def display_formal_time_created(self):
        # http://tools.ietf.org/html/rfc2822.html#section-3.3
        return self.created.strftime("%d %b %Y, %H:%M:%S")

    def inspect(self):
        raise NotImplementedError()

    def display_inspect(self):
        return json.dumps(self.inspect().response, indent=2)

    def __eq__(self, other):
        return type(self) == type(other) and self._id == other._id

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._id)


def graceful_chain_get(d, *args):
    if not d:
        return None
    t = copy.deepcopy(d)
    for arg in args:
        try:
            t = t[arg]
        except (AttributeError, KeyError, TypeError):
            return None
    return t


class DockerImage(DockerObject):
    @property
    def image_id(self):
        if self._id is None:
            self.set_id()
        return self._id

    @property
    def parent_id(self):
        if self.data:
            return self.data.get("ParentId", None)
        else:
            inspect_operation = self.inspect(cached=True)
            return inspect_operation.response.get("Parent", None)

    @property
    def pretty_object_type(self):
        return "Image"

    @property
    def parent_image(self):
        try:
            parent_id = self.parent_id
        except Exception as ex:
            logger.error("error while getting parent ID of image %s: %r", self, ex)
            logger.info(traceback.format_exc())
            raise
        if parent_id:
            return self.docker_backend.get_image_by_id(parent_id)
        else:
            return self.docker_backend.scratch_image

    @property
    def children(self):
        return self.docker_backend.get_images_for_parent(self)

    def get_next_sibling(self):
        imgs = self.parent_image.children
        if len(imgs) == 1:
            return None
        try:
            return imgs[imgs.index(self) + 1]
        except IndexError:
            return None

    def get_prev_sibling(self):
        imgs = self.parent_image.children
        if len(imgs) == 1:
            return None
        # 0 - 1 turns into -1 which turns into last element which creates cycle
        # which totally messes up whole tree
        prev_index = imgs.index(self) - 1
        if prev_index < 0:
            return None
        else:
            return imgs[prev_index]

    @property
    def command(self):
        cmd = graceful_chain_get(self.inspect(cached=True).response, "Config", "Cmd")
        if cmd:
            return " ".join(cmd)
        return ""

    @property
    def container_command(self):
        inspect = self.inspect(cached=True).response
        cmd = graceful_chain_get(inspect, "ContainerConfig", "Cmd")
        if cmd:
            return " ".join(cmd)
        return ""

    @property
    def size(self):
        """
        Size of all layers in bytes

        :return: int
        """
        return self.data["VirtualSize"]

    @property
    def labels(self):
        labels = self.data["Labels"]
        return labels

    @property
    def names(self):
        if self._names is None:
            self._names = []
            if self.data is None:
                return self._names
            for t in self.data["RepoTags"]:
                image_name = ImageNameStruct.parse(t)
                if image_name.to_str():
                    self._names.append(image_name)
            # sort by name length
            self._names.sort(key=lambda x: len(x.to_str()))
        return self._names

    @property
    def short_name(self):
        try:
            ins = self.names[0]
        except IndexError:
            return self.short_id
        if ins.repo == "<none>":
            return self.short_id
        return ins.to_str()

    def base_image(self):
        child_image = self
        while True:
            try:
                parent_image = self.docker_backend.get_image_by_id(child_image.parent_id)
            except Exception as ex:
                logger.warning("error while getting image by ID: %r", ex)
                parent_image = None
            if parent_image is None:
                try:
                    child_image = child_image.parent_image
                except Exception as ex:
                    logger.error("error while getting parent image for image %s: %r", self, ex)
                    return None
                if child_image is None:
                    return None
            else:
                return parent_image

    @operation("Inspect image {object_short_name}.")
    def inspect(self, cached=False):
        if self._inspect is None or cached is False:
            self._inspect = self.d.inspect_image(self.image_id)
        return self._inspect

    @operation("{object_type} {object_short_name} removed!")
    def remove(self):
        return self.d.remove_image(self.image_id)

    @operation("tag of {object_type} {object_short_name} removed!")
    def remove_tag(self, tag):
        assert tag in self.names
        return self.d.remove_image(str(tag))

    def matches_search(self, s):
        return s in self.image_id or \
            s in self.short_name

    def __str__(self):
        if self.names:
            return "{} ({}) {}".format(self.short_id, ", ".join([x.to_str() for x in self.names]), self.container_command)
        else:
            return "{} {}".format(self.short_id, self.container_command)

    def __repr__(self):
        return self.__str__()

    def containers(self):
        return self.docker_backend.get_containers_for_image(self.image_id)


class RootImage(DockerImage):
    """
    this is essentially "scratch" but you cannot inspect it anymore
    """
    def __init__(self, docker_backend):
        self.image_name = "scratch"
        super().__init__(None, docker_backend, object_id="")

    @property
    def parent_id(self):
        return None

    @property
    def parent_image(self):
        return None

    def get_next_sibling(self):
        return None

    def get_prev_sibling(self):
        return None

    @property
    def names(self):
        return [ImageNameStruct.parse(self.image_name)]

    def __str__(self):
        return self.image_name


class DockerContainer(DockerObject):
    """
    Container related logic
    """

    def __str__(self):
        return "{} ({})".format(self.container_id, self.short_name)

    # properties

    @property
    def container_id(self):
        if self._id is None:
            self.set_id()
        return self._id

    @property
    def names(self):
        if self._names is None:
            self._names = []
            for t in self.data["Names"] or []:
                self._names.append(t)
            # sort by name length
            self._names.sort(key=lambda x: len(x))
        return self._names

    @property
    def command(self):
        return self.data["Command"]

    @property
    def status(self):
        return self.data["Status"]

    @property
    def running(self):
        return self.status.startswith("Up")

    @property
    def short_name(self):
        try:
            return self.names[0]
        except IndexError:
            return self.short_id

    @property
    def pretty_object_type(self):
        return "Container"

    @property
    def image_id(self):
        """ this container is created from image with id..."""
        try:
            # docker >= 1.9
            image_id = self.data["ImageID"]
        except KeyError:
            # docker <= 1.8
            image_id = self.inspect(cached=True).response["Image"]
        return image_id

    # methods

    def image_name(self):
        image = self.docker_backend.get_image_by_id(self.image_id)
        if image is not None:
            return image.short_name
        else:
            return self.image_id[:12]

    def matches_search(self, s):
        return s in self.container_id or \
               s in self.short_name
    # api calls

    @operation("Inspect container {object_short_name}.")
    def inspect(self, cached=False):
        if cached is False or self._inspect is None:
            self._inspect = self.d.inspect_container(self.container_id)
        return self._inspect

    @operation("Logs of container {object_short_name} received.")
    def logs(self, follow=False):
        # when tail is set to all, it takes ages to populate widget
        logs_data = self.d.logs(self.container_id, stream=follow, tail=16)
        return logs_data

    @operation("{object_type} {object_short_name} removed!")
    def remove(self):
        self.d.remove_container(self.container_id)

    @operation("{object_type} {object_short_name} started.")
    def start(self):
        self.d.start(self.container_id)

    @operation("{object_type} {object_short_name} stopped.")
    def stop(self):
        self.d.stop(self.container_id)

    @operation("{object_type} {object_short_name} restarted.")
    def restart(self):
        self.d.restart(self.container_id)

    @operation("{object_type} {object_short_name} killed.")
    def kill(self):
        self.d.kill(self.container_id)

    @operation("{object_type} {object_short_name} paused.")
    def pause(self):
        self.d.pause(self.container_id)

    @operation("{object_type} {object_short_name} unpaused.")
    def unpause(self):
        self.d.unpause(self.container_id)


class DockerBackend:
    """
    backend for docker
    """

    def __init__(self):
        self._containers = None
        self._images = None  # displayed images
        self._all_images = None  # docker images -a
        kwargs = docker.utils.kwargs_from_env(assert_hostname=False)
        try:
            self.client = docker.AutoVersionClient(**kwargs)
        except docker.errors.DockerException as ex:
            raise TerminateApplication("can't establish connection to docker daemon: {0}".format(str(ex)))
        self.scratch_image = RootImage(self)

    # backend queries

    @operation("Get list of images.")
    def get_images(self, cached=True):
        if cached is False or self._images is None:
            logger.debug("doing images() query")
            self._images = {}
            for i in self.client.images():
                img = DockerImage(i, self)
                self._images[img.image_id] = img
            self._all_images = {}
            for i in self.client.images(all=True):
                img = DockerImage(i, self)
                self._all_images[img.image_id] = img
        return list(self._images.values())

    @operation("Get list of containers.")
    def get_containers(self, cached=True, stopped=True):
        if cached is False or self._containers is None:
            logger.debug("doing containers() query")
            self._containers = {}
            for c in self.client.containers(all=stopped):
                container = DockerContainer(c, self)
                self._containers[container.container_id] = container
        if not stopped:
            return [x for x in list(self._containers.values()) if x.running]
        return list(self._containers.values())

    def realtime_updates(self):
        for event in self.client.events(decode=True):
            logger.debug("RT event: %s", event)

            # event["from'] means it's a container
            if "from" in event:
                # inspect doesn't contain info about status and you can't query just one
                # container with containers()
                # let's do full-blown containers() query; it's not that expensive
                self.get_containers(cached=False)
            else:
                # similar as ^
                # images() doesn't support query by ID
                # inspect doesn't contain info about repositories
                self.get_images(cached=False)
            content, _, _ = self.filter(containers=True, images=True, stopped=True,
                                        cached=True, sort_by_created=True)
            yield content

    # service methods

    def get_image_by_id(self, image_id):
        return self._all_images.get(image_id)

    def get_images_for_parent(self, image):
        if not image:
            return []
        l = sorted([x for x in self._all_images.values() if x.parent_image == image], key=lambda x: x.created_int)
        return l

    def get_container_by_id(self, container_id):
        return self._containers.get(container_id)

    def get_containers_for_image(self, image_id):
        return [container for container in self._containers.values() if container.image_id == image_id]

    def filter(self, containers=True, images=True, stopped=True, cached=False, sort_by_created=True):
        """
        since django is so awesome, let's use their ORM API

        :return:
        """
        content = []
        containers_o = None
        images_o = None
        # return containers when containers=False and running=True
        if containers or not stopped:
            containers_o = self.get_containers(cached=cached, stopped=stopped)
            content += containers_o.response
        if images:
            images_o = self.get_images(cached=cached)
            content += images_o.response
        if sort_by_created:
            content.sort(key=lambda x: x.created, reverse=True)
        return content, containers_o, images_o



import logging

import urwid

from sen.tui.widgets.list.base import VimMovementListBox
from sen.tui.widgets.list.util import RowWidget

logger = logging.getLogger(__name__)


def assemble_rows(data, headers=None):
    """
    :param data: list of lists:
    [["row 1 column 1", "row 1 column 2"],
     ["row 2 column 1", "row 2 column 2"]]

    each item consists of instance of urwid.Text
    """
    rows = []
    max_lengths = []

    # shitty performance, here we go
    # it would be way better to do a single double loop and provide mutable variable
    for row in data:
        col_index = 0
        for widget in row:
            l = len(widget.text)
            try:
                max_lengths[col_index] = max(l, max_lengths[col_index])
            except IndexError:
                max_lengths.append(l)
            col_index += 1

    if headers:
        header_widgets = []
        for header in headers:
            try:
                header_widgets.append(urwid.Text(*header[0], **header[1]))
            except IndexError:
                header_widgets.append(urwid.Text(*header[0]))
        rows.append(urwid.Columns(header_widgets, dividechars=1))

    for row in data:
        rows.append(
                RowWidget([(max_lengths[idx], item) for idx, item in enumerate(row)])
        )

    return rows


class ResponsiveTable(VimMovementListBox):
    def __init__(self, data, headers=None):
        """
        :param headers: list of widgets which should be displayed as headers
        :param data: list of lists:
            [["row 1 column 1", "row 1 column 2"],
             ["row 2 column 1", "row 2 column 2"]]

             each item consists of instance of urwid.Text
        """

        self.walker = urwid.SimpleFocusListWalker(assemble_rows(data, headers=headers))

        super().__init__(self.walker)





MAIN_LIST_FOCUS = "main_list_focus"

STATUS_BG = "#06a"
STATUS_BG_FOCUS = "#08d"

# name, fg, bg, mono, fg_h, bg_h
PALLETE = [
    (MAIN_LIST_FOCUS, 'default', 'brown', "default", "white", "#060"),  # a60
    ('main_list_lg', 'light gray', 'default', "default", "g100", "default"),
    ('main_list_dg', 'dark gray', 'default', "default", "g78", "default"),
    ('main_list_ddg', 'dark gray', 'default', "default", "g56", "default"),
    ('main_list_white', 'white', 'default', "default", "white", "default"),
    ('main_list_green', 'dark green', 'default', "default", "#0f0", "default"),
    ('main_list_yellow', 'brown', 'default', "default", "#ff0", "default"),
    ('main_list_orange', 'light red', 'default', "default", "#fa0", "default"),
    ('main_list_red', 'dark red', 'default', "default", "#f00", "default"),
    ('image_names', 'light magenta', 'default', "default", "#F0F", "default"),
    ('status_box', 'default', 'black', "default", "g100", STATUS_BG),
    ('status_box_focus', 'default', 'black', "default", "white", STATUS_BG_FOCUS),
    ('status', 'default', 'default', "default", "default", STATUS_BG),
    ('status_text', 'default', 'default', "default", "g100", STATUS_BG),
    ('status_text_green', 'default', 'default', "default", "#0f0", STATUS_BG),
    ('status_text_yellow', 'default', 'default', "default", "#ff0", STATUS_BG),
    ('status_text_orange', 'default', 'default', "default", "#f80", STATUS_BG),
    ('status_text_red', 'default', 'default', "default", "#f66", STATUS_BG),
    ('notif_error', "white", 'dark red', "default", "white", "#f00",),
    ('notif_info', 'white', 'default', "default", "g100", "default"),
    ('notif_important', 'white', 'default', "default", "white", "default"),
    ('notif_text_green', 'white', 'default', "white", "#0f0", "default"),
    ('notif_text_yellow', 'white', 'default', "white", "#ff0", "default"),
    ('notif_text_orange', 'white', 'default', "white", "#f80", "default"),
    ('notif_text_red', 'white', 'default', "white", "#f66", "default"),
    ('tree', 'dark green', 'default', "default", "dark green", "default"),
]

STATUS_BAR_REFRESH_SECONDS = 5
CLEAR_NOTIF_BAR_MESSAGE_IN = 5

HELP_TEXT = """\
# Keybindings

Since I am a heavy `vim` user, these keybindings are trying to stay close to vim.

## Global

/      search (provide empty query to disable searching)
n      next search occurrence
N      previous search occurrence
f4     display only lines matching provided query (provide empty query to clear filtering)
        * main listing provides additional filtering (for more info, check Listing Section)
        * example query: "fed" - display lines containing string "fed"
f5     open a tree view of all images (`docker images --tree` equivalent)
ctrl o next buffer
ctrl i previous buffer
x      remove buffer
ctrl l redraw user interface
h, ?   show help

## Movement

gg     go to first item
G      go to last item
j      go one line down
k      go one line up
pg up
ctrl u go 10 lines up
pg down
ctrl d go 10 lines down

## Listing

@      refresh listing
f4     display only lines matching provided query (provide empty query to clear filtering)
        * space-separated list of query strings, currently supported filters are:
           * t[ype]=c[ontainer[s]]
           * t[ype]=i[mage[s]]
           * s[tate]=r[unning])
          example query may be:
           * "type=container" - show only containers (short equivalent is "t=c")
           * "type=image fedora" - show images with string "fedora" in name (equivalent "t=i fedora")

## Image commands in listing

i      inspect image
d      remove image (irreversible!)
enter  display detailed info about image (when layer is focused)

## Container commands in listing

i      inspect container
l      display logs of container
f      follow logs of container
d      remove container (irreversible!)
t      stop container
s      start container
r      restart container
p      pause container
u      unpause container
X      kill container

## Tree buffer

enter  display detailed info about image (opens image info buffer)

## Image info buffer

d      remove image tag (when image name is focused)
enter  display detailed info about image (when layer is focused)
i      inspect image (when layer is focused)
"""



import logging

import urwid

from sen.docker_backend import RootImage
from sen.tui.constants import MAIN_LIST_FOCUS

logger = logging.getLogger(__name__)


class AdHocAttrMap(urwid.AttrMap):
    """
    Ad-hoc attr map change

    taken from https://github.com/pazz/alot/
    """
    def __init__(self, w, maps, init_map='normal'):
        self.maps = maps
        urwid.AttrMap.__init__(self, w, maps[init_map])
        if isinstance(w, urwid.Text):
            self.attrs = [x[0] for x in self.original_widget.get_text()[1]]

    def set_map(self, attrstring):
        attr_map = {None: self.maps[attrstring]}

        # for urwid.Text only: do hovering for all markups in the widget
        if isinstance(self.original_widget, urwid.Text):
            if attrstring == "normal":
                for a in self.attrs:
                    attr_map[self.maps["focus"]] = a
            elif attrstring == "focus":
                for a in self.attrs:
                    attr_map[a] = self.maps["focus"]
        self.set_attr_map(attr_map)


def get_basic_image_markup(docker_image):
    if isinstance(docker_image, RootImage):
        return [str(docker_image)]

    text_markup = [docker_image.short_id]

    if docker_image.names:
        text_markup.append(" ")
        text_markup.append(("main_list_lg", docker_image.names[0].to_str()))

    text_markup.append(" ")
    text_markup.append(("main_list_ddg", docker_image.container_command))

    return text_markup



import logging
from concurrent.futures.thread import ThreadPoolExecutor

from sen.exceptions import NotifyError
from sen.tui.commands import search, filter
from sen.tui.statusbar import Footer
from sen.tui.buffer import LogsBuffer, MainListBuffer, InspectBuffer, HelpBuffer, ImageInfoBuffer, TreeBuffer
from sen.tui.constants import PALLETE
from sen.docker_backend import DockerBackend

import urwid

from sen.util import log_traceback

logger = logging.getLogger(__name__)


class UI(urwid.MainLoop):
    def __init__(self):
        self.d = DockerBackend()

        # root widget
        self.mainframe = urwid.Frame(urwid.SolidFill())
        self.buffers = []
        self.footer = Footer(self)

        self.executor = ThreadPoolExecutor(max_workers=4)

        root_widget = urwid.AttrMap(self.mainframe, "root")
        self.main_list_buffer = None  # singleton

        screen = urwid.raw_display.Screen()
        screen.set_terminal_properties(256)
        screen.register_palette(PALLETE)

        super().__init__(root_widget, screen=screen)
        self.handle_mouse = False
        self.current_buffer = None

    def run_in_background(self, task, *args, **kwargs):
        logger.info("running task %r(%s, %s) in background", task, args, kwargs)
        self.executor.submit(task, *args, **kwargs)

    def refresh(self):
        try:
            self.draw_screen()
        except AssertionError:
            logger.warning("application is not running")
            pass

    def _set_main_widget(self, widget, redraw):
        """
        add provided widget to widget list and display it

        :param widget:
        :return:
        """
        self.mainframe.set_body(widget)
        self.reload_footer()
        if redraw:
            logger.debug("redraw main widget")
            self.refresh()

    def display_buffer(self, buffer, redraw=True):
        """
        display provided buffer

        :param buffer: Buffer
        :return:
        """
        self.current_buffer = buffer
        self._set_main_widget(buffer.widget, redraw=redraw)

    def add_and_display_buffer(self, buffer, redraw=True):
        """
        add provided buffer to buffer list and display it

        :param buffer:
        :return:
        """
        if buffer not in self.buffers:
            logger.debug("adding new buffer {!r}".format(buffer))
            self.buffers.append(buffer)
        self.display_buffer(buffer, redraw=redraw)

    def pick_and_display_buffer(self, i):
        """
        pick i-th buffer from list and display it

        :param i: int
        :return: None
        """
        if len(self.buffers) == 1:
            # we don't need to display anything
            # listing is already displayed
            return
        else:
            try:
                self.display_buffer(self.buffers[i])
            except IndexError:
                # i > len
                self.display_buffer(self.buffers[0])

    @property
    def current_buffer_index(self):
        return self.buffers.index(self.current_buffer)

    def remove_current_buffer(self):
        # don't allow removing main_list
        if isinstance(self.current_buffer, MainListBuffer):
            logger.warning("you can't remove main list widget")
            return
        self.buffers.remove(self.current_buffer)
        self.current_buffer.destroy()
        # FIXME: we should display last displayed widget here
        self.display_buffer(self.buffers[0], True)

    def unhandled_input(self, key):
        logger.debug("unhandled input: %r", key)
        try:
            if key in ('q', 'Q'):
                self.executor.shutdown(wait=False)
                raise urwid.ExitMainLoop()
            elif key == "ctrl o":
                self.pick_and_display_buffer(self.current_buffer_index - 1)
            elif key == "ctrl i":
                self.pick_and_display_buffer(self.current_buffer_index + 1)
            elif key == "x":
                self.remove_current_buffer()
            elif key == "/":
                self.prompt("/", search)
            elif key == "f4":
                self.footer.prompt("filter ", filter)
            elif key == "n":
                self.current_buffer.find_next()
            elif key == "N":
                self.current_buffer.find_previous()
            elif key in ["h", "?"]:
                self.display_help()
            elif key == "f5":
                self.display_tree()
        except NotifyError as ex:
            self.notify_message(str(ex), level="error")
            logger.error(repr(ex))

    def run(self):
        self.main_list_buffer = MainListBuffer(self.d, self)

        @log_traceback
        def chain_fcs():
            self.main_list_buffer.refresh(focus_on_top=True)
            self.add_and_display_buffer(self.main_list_buffer, redraw=True)
        self.run_in_background(chain_fcs)
        super().run()

    def display_logs(self, docker_container):
        self.add_and_display_buffer(LogsBuffer(docker_container, self))

    def display_and_follow_logs(self, docker_container):
        self.add_and_display_buffer(LogsBuffer(docker_container, self, follow=True))

    def inspect(self, docker_object):
        self.add_and_display_buffer(InspectBuffer(docker_object))

    def display_image_info(self, docker_image):
        try:
            self.add_and_display_buffer(ImageInfoBuffer(docker_image, self))
        except NotifyError as ex:
            self.notify_message(str(ex), level="error")
            logger.error(repr(ex))

    def refresh_main_buffer(self, refresh_buffer=True):
        assert self.main_list_buffer is not None
        if refresh_buffer:
            self.main_list_buffer.refresh()
        self.display_buffer(self.main_list_buffer)

    def display_help(self):
        self.add_and_display_buffer(HelpBuffer())

    def display_tree(self):
        self.add_and_display_buffer(TreeBuffer(self.d, self))

    # FOOTER

    def set_footer(self, widget):
        self.mainframe.set_footer(widget)

    def reload_footer(self):
        self.footer.reload_footer()

    def remove_notification_message(self, message):
        self.footer.remove_notification_message(message)

    def notify_widget(self, *args, **kwargs):
        self.footer.notify_widget(*args, **kwargs)

    def notify_message(self, *args, **kwargs):
        self.footer.notify_message(*args, **kwargs)

    def prompt(self, *args, **kwargs):
        self.footer.prompt(*args, **kwargs)


from flexmock import flexmock
import docker


image_data = {
    'Created': 1414577076,
    'Id': '3ab9a7ed8a169ab89b09fb3e12a14a390d3c662703b65b4541c0c7bde0ee97eb',
    'ParentId': 'a79ad4dac406fcf85b9c7315fe08de5b620c1f7a12f45c8185c843f4b4a49c4e',
    'RepoTags': ['image:latest'],
    'Size': 0,
    'VirtualSize': 856564160
}


version_data = {'ApiVersion': '1.21',
    'Arch': 'amd64',
    'BuildTime': 'Thu Sep 10 17:53:19 UTC 2015',
    'GitCommit': 'af9b534-dirty',
    'GoVersion': 'go1.5.1',
    'KernelVersion': '4.2.5-300.fc23.x86_64',
    'Os': 'linux',
    'Version': '1.9.0-dev-fc24'
}


def images_response(*args, **kwargs):
    global image_data
    return [image_data]


def mock():
    flexmock(docker.Client, images=images_response)
    flexmock(docker.Client, version=lambda *args, **kwargs: version_data)



from sen.docker_backend import DockerBackend
from .real import image_data, mock


def test_images_call():
    mock()
    b = DockerBackend()
    operation = b.get_images()
    images_response = operation.response
    assert len(images_response) == 1
    assert images_response[0].image_id == image_data["Id"]
    assert images_response[0].short_name == image_data["RepoTags"][0]



import logging

from sen.constants import FALLBACK_LOG_PATH

__version__ = "0.1"


def set_logging(name="sen", level=logging.DEBUG, path=FALLBACK_LOG_PATH):
    logger = logging.getLogger(name)
    logger.setLevel(level)

    handler = logging.FileHandler(path)
    handler.setLevel(logging.DEBUG)
    # handler = logging.StreamHandler(sys.stderr)

    formatter = logging.Formatter('%(asctime)s %(name)-18s %(levelname)-6s %(message)s',
                                  '%H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)



"""
Info widgets:
 * display detailed info about an object
"""
import logging

import urwid

from sen.tui.widgets.list.util import get_map, RowWidget
from sen.tui.widgets.table import ResponsiveTable, assemble_rows
from sen.util import humanize_bytes
from sen.tui.widgets.list.base import VimMovementListBox
from sen.tui.widgets.util import AdHocAttrMap, get_basic_image_markup

logger = logging.getLogger(__name__)


class SelectableText(AdHocAttrMap):
    def __init__(self, text, maps=None):
        maps = maps or get_map()
        super().__init__(urwid.Text(text, align="left", wrap="clip"), maps)

    @property
    def text(self):
        return self._original_widget.text

    def keypress(self, size, key):
        """ get rid of tback: `AttributeError: 'Text' object has no attribute 'keypress'` """
        return key


class LayerWidget(SelectableText):
    def __init__(self, ui, docker_image, index=0):
        self.ui = ui
        self.docker_image = docker_image
        separator = "└─"
        if index == 0:
            label = [separator]
        else:
            label = [2 * index * " " + separator]
        super().__init__(label + get_basic_image_markup(docker_image))

    def keypress(self, size, key):
        logger.debug("%s %s %s", self.__class__, key, size)
        if key == "enter":
            self.ui.display_image_info(self.docker_image)
            return
        elif key == "i":
            self.ui.inspect(self.docker_image)  # FIXME: do this async
            return
        return key


class TagWidget(SelectableText):
    """
    so we can easily access image and tag
    """
    def __init__(self, docker_image, tag):
        self.docker_image = docker_image
        self.tag = tag
        super().__init__(str(self.tag))


class ImageInfoWidget(VimMovementListBox):
    """
    display info about image
    """
    def __init__(self, ui, docker_image):
        self.ui = ui
        self.docker_image = docker_image

        self.walker = urwid.SimpleFocusListWalker([])

        # self.widgets = []

        self._basic_data()
        self._containers()
        self._image_names()
        self._layers()
        self._labels()

        super().__init__(self.walker)

        self.set_focus(0)  # or assemble list first and then stuff it into walker

    def _basic_data(self):
        data = [
            [SelectableText("Id", maps=get_map("main_list_green")),
             SelectableText(self.docker_image.image_id)],
            [SelectableText("Created", maps=get_map("main_list_green")),
             SelectableText("{0}, {1}".format(self.docker_image.display_formal_time_created(),
                                              self.docker_image.display_time_created()))],
            [SelectableText("Size", maps=get_map("main_list_green")),
             SelectableText(humanize_bytes(self.docker_image.size))],
            [SelectableText("Command", maps=get_map("main_list_green")),
             SelectableText(self.docker_image.container_command)],
        ]
        self.walker.extend(assemble_rows(data))

    def _image_names(self):
        if not self.docker_image.names:
            return
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Image Names", maps=get_map("main_list_white"))]))
        for n in self.docker_image.names:
            self.walker.append(RowWidget([TagWidget(self.docker_image, n)]))

    def _layers(self):
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Layers", maps=get_map("main_list_white"))]))

        i = self.docker_image
        index = 0
        self.walker.append(RowWidget([LayerWidget(self.ui, self.docker_image, index=index)]))
        while True:
            index += 1
            parent = i.parent_image
            if parent:
                self.walker.append(RowWidget([LayerWidget(self.ui, parent, index=index)]))
                i = parent
            else:
                break

    def _labels(self):
        if not self.docker_image.labels:
            return []
        data = []
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Labels", maps=get_map("main_list_white"))]))
        for label_key, label_value in self.docker_image.labels.items():
            data.append([SelectableText(label_key, maps=get_map("main_list_green")), SelectableText(label_value)])
        self.walker.extend(assemble_rows(data))

    def _containers(self):
        if not self.docker_image.containers():
            return
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Containers", maps=get_map("main_list_white"))]))
        for container in self.docker_image.containers():
            self.walker.append(RowWidget([SelectableText(str(container))]))

    def keypress(self, size, key):
        logger.debug("%s, %s", key, size)

        if key == "d":
            img = self.focus.columns.widget_list[0].docker_image
            tag = self.focus.columns.widget_list[0].tag
            img.remove_tag(tag)  # FIXME: do this async
            self.walker.remove(self.focus)
            return

        key = super().keypress(size, key)
        return key



import logging

from sen.docker_backend import DockerContainer, RootImage
from sen.exceptions import NotifyError
from sen.tui.constants import HELP_TEXT
from sen.tui.widgets.info import ImageInfoWidget
from sen.tui.widgets.list.main import MainListBox
from sen.tui.widgets.list.util import get_operation_notify_widget
from sen.tui.widgets.list.common import AsyncScrollableListBox, ScrollableListBox
from sen.tui.widgets.tree import ImageTree


logger = logging.getLogger(__name__)


class Buffer:
    """
    base buffer class
    """
    display_name = None  # display in status bar
    widget = None  # display this in main frame
    tag = None  # single char, for status

    def __init__(self):
        logger.debug("creating buffer %r", self)

    def __repr__(self):
        return "{}(name={!r}, widget={!r})".format(
            self.__class__.__name__, self.display_name, self.widget)

    def destroy(self):
        destroy_method = getattr(self.widget, "destroy", None)
        if destroy_method:
            destroy_method()

    def find_previous(self, s=None):
        logger.debug("searching next %r in %r", s, self.__class__.__name__)
        try:
            self.widget.find_previous(s)
        except AttributeError as ex:
            logger.debug(repr(ex))
            raise NotifyError("Can't search in this buffer.")

    def find_next(self, s=None):
        logger.debug("searching next %r in %r", s, self.__class__.__name__)
        try:
            self.widget.find_next(s)
        except AttributeError as ex:
            logger.debug(repr(ex))
            raise NotifyError("Can't search in this buffer.")

    def build_status_bar(self):
        status_bar = getattr(self.widget, "status_bar", None)
        if status_bar:
            return status_bar()

    def filter(self, s):
        logger.debug("filter widget %r with query %r", self.widget, s)
        widgets = self.widget.filter(s)
        if widgets is not None:
            self.widget.set_body(widgets)


class ImageInfoBuffer(Buffer):
    tag = "I"

    def __init__(self, docker_image, ui):
        """
        :param docker_image:
        :param ui: ui object so we refresh
        """
        if isinstance(docker_image, RootImage):
            raise NotifyError("Image \"scratch\" doesn't provide any more information.")
        self.display_name = docker_image.short_name
        self.widget = ImageInfoWidget(ui, docker_image)
        super().__init__()


class TreeBuffer(Buffer):
    display_name = "Tree"
    tag = "T"

    def __init__(self, docker_backend, ui):
        """
        """
        self.widget = ImageTree(docker_backend, ui)
        super().__init__()


class MainListBuffer(Buffer):
    display_name = "Listing"
    tag = "M"

    def __init__(self, docker_backend, ui):
        self.ui = ui
        self.widget = MainListBox(docker_backend, ui)
        super().__init__()

    def refresh(self, focus_on_top=False):
        self.widget.populate(focus_on_top=focus_on_top)
        self.ui.refresh()


class LogsBuffer(Buffer):

    def __init__(self, docker_object, ui, follow=False):
        """

        :param docker_object: container to display logs
        :param ui: ui object so we refresh
        """
        self.tag = "F" if follow else "L"
        self.display_name = docker_object.short_name
        if isinstance(docker_object, DockerContainer):
            try:
                pre_message = "Getting logs for container {}...".format(docker_object.short_name)
                ui.notify_message(pre_message)
                operation = docker_object.logs(follow=follow)
                ui.remove_notification_message(pre_message)
                ui.notify_widget(get_operation_notify_widget(operation, display_always=False))
                if follow:
                    self.widget = AsyncScrollableListBox(operation.response, ui)
                else:
                    self.widget = ScrollableListBox(operation.response)
            except Exception as ex:
                # FIXME: let's catch 404 and print that container doesn't exist
                #        instead of printing ugly HTTP error
                raise NotifyError("Error getting logs for container %s: %r" % (docker_object, ex))
        else:
            raise NotifyError("Only containers have logs.")
        super().__init__()


class InspectBuffer(Buffer):
    tag = "I"

    def __init__(self, docker_object):
        """

        :param docker_object: object to inspect
        """
        self.display_name = docker_object.short_name
        inspect_data = docker_object.display_inspect()
        self.widget = ScrollableListBox(inspect_data)
        self.display_name = docker_object.short_name
        super().__init__()


class HelpBuffer(Buffer):
    tag = "H"
    display_name = ""

    def __init__(self):
        """
        """
        self.widget = ScrollableListBox(HELP_TEXT)
        super().__init__()



import logging
import threading

import urwid

from sen.tui.widgets.list.base import VimMovementListBox
from sen.util import _ensure_unicode


logger = logging.getLogger(__name__)


class ScrollableListBox(VimMovementListBox):
    def __init__(self, text):
        text = _ensure_unicode(text)
        list_of_texts = text.split("\n")
        self.walker = urwid.SimpleFocusListWalker([
            urwid.AttrMap(urwid.Text(t, align="left", wrap="any"), "main_list_dg", "main_list_white")
            for t in list_of_texts
        ])
        super().__init__(self.walker)


class AsyncScrollableListBox(VimMovementListBox):
    def __init__(self, generator, ui, static_data=None):
        self.log_texts = []
        if static_data:
            static_data = _ensure_unicode(static_data).split("\n")
            for d in static_data:
                log_entry = d.strip()
                if log_entry:
                    self.log_texts.append(urwid.Text(("main_list_dg", log_entry),
                                                     align="left", wrap="any"))
        walker = urwid.SimpleFocusListWalker(self.log_texts)
        super(AsyncScrollableListBox, self).__init__(walker)

        def fetch_logs():
            for line in generator:
                line = _ensure_unicode(line)
                if self.stop.is_set():
                    break
                if self.filter_query:
                    if self.filter_query not in line:
                        continue
                walker.append(
                    urwid.AttrMap(
                        urwid.Text(line.strip(), align="left", wrap="any"), "main_list_dg", "main_list_white"
                    )
                )
                walker.set_focus(len(walker) - 1)
                ui.refresh()

        self.stop = threading.Event()
        self.thread = threading.Thread(target=fetch_logs, daemon=True)
        self.thread.start()

    def destroy(self):
        self.stop.set()




# -*- coding: utf-8 -*-
from concurrent.futures.thread import ThreadPoolExecutor
import time

from sen.util import _ensure_unicode, log_traceback

import pytest


@pytest.mark.parametrize("inp,expected", [
    ("a", "a"),
    (b"a", "a"),
    ("\u2606", "☆"),
    (b'\xe2\x98\x86', "☆"),
])
def test_ensure_unicode(inp, expected):
    assert _ensure_unicode(inp) == expected


def test_log_traceback(caplog):
    @log_traceback
    def f():
        raise Exception()
    f()
    assert caplog.records()[0].message.startswith("starting thread")
    assert caplog.records()[1].message.startswith("Traceback")
    assert caplog.records()[1].message.endswith("Exception\n")


def test_log_traceback_without_tb(caplog):
    @log_traceback
    def f():
        pass
    f()
    assert caplog.records()[0].message.startswith("starting thread")
    assert caplog.records()[1].message.startswith("closing thread")


def test_log_traceback_threaded(caplog):
    @log_traceback
    def f():
        raise Exception()

    e = ThreadPoolExecutor(max_workers=1)
    f = e.submit(f)
    while f.running():
        time.sleep(0.1)

    assert caplog.records()[0].message.startswith("starting thread")
    assert caplog.records()[1].message.startswith("Traceback")
    assert caplog.records()[1].message.endswith("Exception\n")



import datetime
import logging

import urwid

from sen.tui.constants import MAIN_LIST_FOCUS


logger = logging.getLogger(__name__)


def get_color_text(markup, color_attr="status_text"):
    w = urwid.AttrMap(urwid.Text(markup), color_attr)
    return len(markup), w


def get_operation_notify_widget(operation, notif_level="info", display_always=True):
    attr = "notif_{}".format(notif_level)
    took = operation.took
    text_list = []
    if took > 300:
        fmt_str = "{} Query took "
        text_list.append((attr, fmt_str.format(operation.pretty_message)))
        command_took_str = "{:.2f}".format(took)
        if took < 500:
            text_list.append(("notif_text_yellow", command_took_str))
        elif took < 1000:
            text_list.append(("notif_text_orange", command_took_str))
        else:
            command_took_str = "{:.2f}".format(took / 1000.0)
            text_list.append(("notif_text_red", command_took_str))
            text_list.append((attr, " s"))
        if took < 1000:
            text_list.append((attr, " ms"))
    elif display_always:
        text_list.append((attr, operation.pretty_message))
    else:
        return
    return urwid.AttrMap(urwid.Text(text_list), attr)


def get_map(defult="main_list_dg"):
    return {"normal": defult, "focus": MAIN_LIST_FOCUS}


def get_time_attr_map(t):
    """
                                                       now -> |
                            hour ago -> |
        day ago -> |
    |--------------|--------------------|---------------------|
    """
    now = datetime.datetime.now()
    if t + datetime.timedelta(hours=3) > now:
        return get_map("main_list_white")
    if t + datetime.timedelta(days=3) > now:
        return get_map("main_list_lg")
    else:
        return get_map("main_list_dg")


class RowWidget(urwid.AttrMap):
    def __init__(self, columns, attr="main_list_dg", focus_map=MAIN_LIST_FOCUS):
        self.columns = urwid.Columns(columns, dividechars=1)
        super().__init__(self.columns, attr, focus_map=focus_map)

    def selectable(self):
        return True

    def render(self, size, focus=False):
        for w in self.columns.widget_list:
            w.set_map('focus' if focus else 'normal')
        return urwid.AttrMap.render(self, size, focus)



#!/usr/bin/env python3
"""
yes, this is python 3 ONLY project
"""

from __future__ import print_function

import sys

# let's be so gentle and print the error message even on python 2
# running on py 2, halt
if sys.version_info.major <= 2:
    error_message = """\
It looks like you are running sen with python 2. I'm sorry but sen is a python 3 only project.
Please see installation steps at official project page:

https://github.com/TomasTomecek/sen"""
    print(error_message, file=sys.stderr)
    sys.exit(2)

import argparse
import logging

from sen import set_logging
from sen.exceptions import TerminateApplication
from sen.tui.init import UI
from sen.util import get_log_file_path


logger = logging.getLogger("sen")


def main():
    parser = argparse.ArgumentParser(
        description="Terminal User Interface for Docker Engine"
    )
    exclusive_group = parser.add_mutually_exclusive_group()
    exclusive_group.add_argument("--debug", action="store_true", default=None)

    args = parser.parse_args()

    # if args.debug:
    set_logging(level=logging.DEBUG, path=get_log_file_path())
    # else:
    #     set_logging(level=logging.INFO, path=setup_dirs())

    logger.info("application started")

    try:
        ui = UI()
    except TerminateApplication as ex:
        print("Error: {0}".format(str(ex)), file=sys.stderr)
        return 1

    try:
        ui.run()
    except KeyboardInterrupt:
        print("Quitting on user request.")
        return 1
    except Exception as ex:  # pylint: disable=broad-except
        if args.debug:
            raise
        else:
            logger.error("Exception caught: %r", ex)
            return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



import logging

import urwid
from sen.exceptions import NotifyError
from sen.util import log_traceback

logger = logging.getLogger(__name__)


@log_traceback
def search(ui, oldfooter, edit_widget, text_input):
    logger.debug("%r %r", edit_widget, text_input)
    if text_input.endswith("\n"):
        # TODO: implement incsearch
        #   - match needs to be highlighted somehow, not with focus though
        ui.footer.prompt_bar = None
        ui.mainframe.set_footer(oldfooter)
        try:
            ui.current_buffer.find_next(text_input[:-1])
        except NotifyError as ex:
            logger.error(repr(ex))
            ui.notify_message(str(ex), level="error")
        ui.mainframe.set_focus("body")
        ui.reload_footer()


@log_traceback
def filter(ui, oldfooter, edit_widget, text_input):
    logger.debug("%r %r", edit_widget, text_input)
    if text_input.endswith("\n"):
        ui.footer.prompt_bar = None
        ui.mainframe.set_footer(oldfooter)
        try:
            ui.current_buffer.filter(text_input[:-1])
        except NotifyError as ex:
            logger.error(repr(ex))
            ui.notify_message(str(ex), level="error")
        ui.mainframe.set_focus("body")
        ui.reload_footer()



import os
import logging
import functools
import traceback

from sen.constants import PROJECT_NAME, LOG_FILE_NAME

logger = logging.getLogger(__name__)


def _ensure_unicode(s):
    try:
        return s.decode("utf-8")
    except AttributeError:
        return s


def log_traceback(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger.info("starting thread for function %s", func)
        try:
            response = func(*args, **kwargs)
        except Exception:
            logger.error(traceback.format_exc())
        else:
            logger.info("closing thread for function %s", func)
            return response
    return wrapper


def setup_dirs():
    try:
        top_dir = os.path.abspath(os.path.expanduser(os.environ["XDG_CACHE_HOME"]))
    except KeyError:
        top_dir = os.path.abspath(os.path.expanduser("~/.cache"))
    our_cache_dir = os.path.join(top_dir, PROJECT_NAME)
    os.makedirs(our_cache_dir, mode=0o775, exist_ok=True)
    return our_cache_dir


def get_log_file_path():
    return os.path.join(setup_dirs(), LOG_FILE_NAME)


def humanize_bytes(bytesize, precision=2):
    """
    Humanize byte size figures

    https://gist.github.com/moird/3684595
    """
    abbrevs = (
        (1 << 50, 'PB'),
        (1 << 40, 'TB'),
        (1 << 30, 'GB'),
        (1 << 20, 'MB'),
        (1 << 10, 'kB'),
        (1, 'bytes')
    )
    if bytesize == 1:
        return '1 byte'
    for factor, suffix in abbrevs:
        if bytesize >= factor:
            break
    if factor == 1:
        precision = 0
    return '%.*f %s' % (precision, bytesize / float(factor), suffix)


# from concurrent.futures.thread import ThreadPoolExecutor
import logging
import threading

import urwid
from sen.tui.constants import CLEAR_NOTIF_BAR_MESSAGE_IN
from sen.util import log_traceback

logger = logging.getLogger(__name__)


class Footer:
    """
    wrapper class for all statusbar/notifbar related methods
    """

    def __init__(self, ui):
        self.ui = ui
        self.notif_bar = None
        self.status_bar = self.build_statusbar()
        self.prompt_bar = None
        # self.executor = ThreadPoolExecutor(max_workers=8)
        self.notifications = []
        # lock when managing notifications:
        #  * when accessing self.notifications
        #  * when accessing widgets
        # and most importantly, remember, locking is voodoo
        self.notifications_lock = threading.Lock()

    def reload_notif_bar(self):
        logger.debug("reload notif bar")
        bottom = []
        if self.notif_bar:
            bottom.append(self.notif_bar)
        if self.prompt_bar:
            bottom.append(self.prompt_bar)
        else:
            bottom.append(self.status_bar)
        self.ui.set_footer(urwid.Pile(bottom))
        self.ui.refresh()

    def reload_footer(self):
        logger.debug("reload footer")
        bottom = []
        if self.notif_bar:
            bottom.append(self.notif_bar)
        if self.prompt_bar:
            bottom.append(self.prompt_bar)
        else:
            self.status_bar = self.build_statusbar()
            bottom.append(self.status_bar)
        self.ui.set_footer(urwid.Pile(bottom))
        # do not refresh here b/c mainloop might not started

    def build_statusbar(self):
        """construct and return statusbar widget"""
        try:
            left_widgets = self.ui.current_buffer.build_status_bar() or []
        except AttributeError:
            left_widgets = []
        text_list = []
        for idx, buffer in enumerate(self.ui.buffers):
            #  #1 [I] fedora #2 [L]
            fmt = "#{idx} [{tag}] {name}"
            markup = fmt.format(idx=idx, tag=buffer.tag, name=buffer.display_name)
            text_list.append((
                "status_box_focus" if buffer == self.ui.current_buffer else "status_box",
                markup,
            ))
            text_list.append(" ")
        text_list = text_list[:-1]

        if text_list:
            buffer_text = urwid.Text(text_list, align="right")
        else:
            buffer_text = urwid.Text("", align="right")
        columns = urwid.Columns(left_widgets + [buffer_text])
        return urwid.AttrMap(columns, "status")

    def prompt(self, prompt_text, callback):
        """
        prompt for text input.
        """
        oldfooter = self.ui.mainframe.get_footer()

        # set up widgets
        leftpart = urwid.Text(prompt_text, align='left')
        editpart = urwid.Edit(multiline=True)

        # build promptwidget
        edit = urwid.Columns(
            [
                ('fixed', len(prompt_text), leftpart),
                ('weight', 1, editpart),
            ])
        self.prompt_bar = urwid.AttrMap(edit, "main_list_dg")

        widgets = []
        if self.notif_bar:
            widgets.append(self.notif_bar)
        widgets.append(self.prompt_bar)
        self.ui.mainframe.set_footer(urwid.Pile(widgets))

        self.ui.mainframe.set_focus("footer")

        urwid.connect_signal(editpart, "change", callback, user_args=[self.ui, oldfooter])

    def remove_notification_message(self, message):
        logger.debug("requested remove of message %r from notif bar", message)
        with self.notifications_lock:
            if message in self.notifications:
                self.notifications.remove(message)
            if self.notif_bar:
                newpile = self.notif_bar.widget_list
                for w in newpile:
                    if w.original_widget.get_text()[0] == message:
                        logger.debug("remove widget %r from new pile", w)
                        newpile.remove(w)
                        break
                if newpile:
                    self.notif_bar = urwid.Pile(newpile)
                else:
                    self.notif_bar = None
            else:
                self.notif_bar = None
        self.reload_notif_bar()

    def notify_message(self, message=None, level="info", clear_if_dupl=True,
                        clear_in=CLEAR_NOTIF_BAR_MESSAGE_IN):
        """
        :param level: str, {info, error}

        opens notification popup.
        """
        with self.notifications_lock:
            if clear_if_dupl:
                if message in self.notifications:
                    # notification is already displayed
                    return
            self.notifications.append(message)
        logger.debug("display notification %r", message)
        widget = urwid.AttrMap(urwid.Text(message), "notif_{}".format(level))
        self.notify_widget(widget, message=message, clear_in=clear_in)

    def notify_widget(self, widget, message=None, clear_in=CLEAR_NOTIF_BAR_MESSAGE_IN):
        """
        opens notification popup.

        :param widget: instance of Widget, widget to display
        :param message: str, message to remove from list of notifications
        :param clear_in: int, time seconds when notification should be removed
        """

        @log_traceback
        def clear_notification(*args, **kwargs):
            # time.sleep(clear_in)
            logger.debug("remove widget %r from notif bar", widget)
            with self.notifications_lock:
                if message in self.notifications:
                    self.notifications.remove(message)
                if self.notif_bar:
                    newpile = self.notif_bar.widget_list
                    if widget in newpile:
                        newpile.remove(widget)
                    if newpile:
                        self.notif_bar = urwid.Pile(newpile)
                    else:
                        self.notif_bar = None
                else:
                    self.notif_bar = None
            self.reload_notif_bar()

        if not widget:
            return

        logger.debug("display notification widget %s", widget)

        widget_list = [widget]
        with self.notifications_lock:
            # stack errors, don't overwrite them
            if not self.notif_bar:
                self.notif_bar = urwid.Pile(widget_list)
            else:
                logger.debug(self.notif_bar.widget_list)
                newpile = self.notif_bar.widget_list + widget_list
                self.notif_bar = urwid.Pile(newpile)

        self.reload_notif_bar()
        # self.executor.submit(clear_notification)
        self.ui.set_alarm_in(clear_in, clear_notification)



#!/usr/bin/env python3

import os

from setuptools import setup, find_packages


def get_requirements():
    this_dir = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(this_dir, 'requirements.txt')
    try:
        with open(path) as f:
            packages = f.read().splitlines()
    except (IOError, OSError) as ex:
        raise RuntimeError("Can't open file with requirements: %r", ex)
    return packages


setup(
    name='sen',
    version='0.2.2',
    description="Terminal User Interface for Docker Engine",
    author='Tomas Tomecek',
    author_email='tomas@tomecek.net',
    url="https://github.com/TomasTomecek/sen/",
    license="MIT",
    entry_points={
        'console_scripts': ['sen=sen.cli:main'],
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=get_requirements(),
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Developers',
                 'Intended Audience :: System Administrators',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3 :: Only',
                 'Programming Language :: Python :: 3.4',
                 'Topic :: System :: Monitoring',
                 ],
)
